#include "Camera.h"

/** Constructor */
Camera::Camera() : mode_(SPHERICAL), rot_(), viewMatrix_(), projectionMatrix_(), normalizedMouseXY_(),
    viewport_(), transX_(0), transY_(0), zoom_(20), moveX_(0), moveZ_(0), position_(0,0,20),
    center_(), up_(0,1,0), theta_(-90), phi_(0), globalScale_(1)
{}

/** Destructor */
Camera::~Camera()
{}

/** Setters/Getters */
void Camera::setViewport(const QSizeF &viewport) { viewport_ = viewport; glViewport(0, 0, viewport_.width(), viewport_.height());}
float Camera::getZoom() const { return zoom_; }
const QSizeF& Camera::getViewport() const { return viewport_; }
const QMatrix4x4& Camera::getViewMatrix() const { return viewMatrix_; }
const QMatrix4x4& Camera::getProjectionMatrix() const { return projectionMatrix_; }
void Camera::setMoveX(int direction) { moveX_ = direction; }
void Camera::setMoveZ(int direction) { moveZ_ = direction; }
void Camera::setGlobalScale(float scale) { globalScale_ = scale; }
bool Camera::isMoving() const { return moveX_!=0 || moveZ_!=0; }
void Camera::setZoom(float zoom) { position_.setZ(zoom); zoom_ = zoom; }
Camera::CameraMode Camera::getMode() const { return mode_; }
void Camera::setMode(CameraMode mode)
{
    rot_ = QQuaternion();
    mode_ = mode;
    float global = globalScale_;
    reset();
    globalScale_ = global;
    zoom(-0.4f);
}

/** Start camera (store mouse coordinates) */
void Camera::start(const QPointF& mouseXY)
{
    normalizedMouseXY_=Geometry::normalizedMouse(mouseXY, viewport_);
}

/** Compute rotation values (by updating the quaternion) */
void Camera::rotate(const QPointF& mouseXY)
{
    QPointF normalizedMouseXY = Geometry::normalizedMouse(mouseXY, viewport_);
    if (mode_==PLANE)
    {
        QLineF diff(normalizedMouseXY_, normalizedMouseXY);
        QVector3D axe = QVector3D(-diff.dy(), diff.dx(), 0.0f).normalized();
        rot_ = QQuaternion::fromAxisAndAngle(axe, diff.length()*180/M_PI)*rot_;
    }
    else if (mode_==SPHERICAL)
    {
        QVector3D mouseOnSphereBefore = Geometry::mouseOnUnitSphere(normalizedMouseXY_);
        QVector3D mouseOnSphereAfter = Geometry::mouseOnUnitSphere(normalizedMouseXY);
        float angle=acosf(QVector3D::dotProduct(mouseOnSphereBefore, mouseOnSphereAfter));
        QVector3D axe = QVector3D::crossProduct(mouseOnSphereBefore, mouseOnSphereAfter);
        rot_ = QQuaternion::fromAxisAndAngle(axe, angle*180/M_PI)*rot_;
    }
    else if (mode_==FREELOOK)
    {
        QLineF diff(normalizedMouseXY_, normalizedMouseXY);
        theta_ += diff.dx()*50.f;
        phi_ += diff.dy()*50.f;
        if(phi_>89.f)
            phi_=89.f;
        else if (phi_<-89.f)
            phi_=-89.f;
        float theta=theta_*M_PI/180.f;
        float phi = phi_*M_PI/180.f;
        QVector3D view(cos(phi)*cos(theta),sin(phi),cos(phi)*sin(theta));
        center_ = position_ + view.normalized();
    }
    normalizedMouseXY_ = normalizedMouseXY;
}

/** Compute translation values */
void Camera::translate(const QPointF& mouseDiff)
{
    transX_-=mouseDiff.x()*globalScale_;
    transY_+=mouseDiff.y()*globalScale_;
}

/** Update model view matrices */
void Camera::updateView()
{
    viewMatrix_.setToIdentity();
    if(mode_==FREELOOK)
        viewMatrix_.lookAt(QVector3D(position_.x(), position_.y(), position_.z()),
                           QVector3D(center_.x(), center_.y(), center_.z()),QVector3D( 0, 1, 0));
    else
    {
        viewMatrix_.lookAt(QVector3D(transX_, transY_, zoom_),QVector3D(transX_, transY_, 0),QVector3D( 0, 1, 0));
        viewMatrix_.rotate(rot_);
    }
}

/** Update projection matrix */
void Camera::updateProjection()
{
    projectionMatrix_.setToIdentity();
    projectionMatrix_.perspective(70, (GLdouble)viewport_.width()/viewport_.height(), 0.01, 100000);
}

/** Zoom */
void Camera::zoom(float delta)
{
    if(mode_==FREELOOK)
    {
        QVector3D view = (center_-position_).normalized();
        QVector3D trans = view*delta*globalScale_;
        position_+=trans;
        center_+=trans;
    }
    else
        zoom_ = std::max(0.00001f, zoom_-delta*globalScale_);
}

/** Update translation */
void Camera::updateTranslation()
{
    if(mode_==FREELOOK)
    {
        QVector3D view = (center_-position_).normalized();
        QVector3D ortho = (QVector3D::crossProduct(view,up_)).normalized();
        QVector3D trans = (ortho*moveX_ - view*moveZ_)*globalScale_/400.f ;
        position_+=trans;
        center_+=trans;
    }
    else
    {
        transX_+=moveX_*globalScale_/400.f;
        zoom_=std::max(0.00001f, zoom_+moveZ_*globalScale_/400.f);
    }
}
/** Load projection matrix (only for immediate mode...) */
void Camera::loadProjectionMatrix()
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    qreal *dataMat = projectionMatrix_.data();
    GLfloat matriceArray[16];
    for (int i= 0; i < 16; ++i)
        matriceArray[i] = dataMat[i];
    glMultMatrixf(matriceArray);
}

/** Load view matrix (only for immediate mode...) */
void Camera::loadViewMatrix()
{
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    qreal *dataMat = viewMatrix_.data();
    GLfloat matriceArray[16];
    for (int i= 0; i < 16; ++i)
        matriceArray[i] = dataMat[i];
    glMultMatrixf(matriceArray);
}

/** Reset camera */
void Camera::reset()
{
    rot_ = QQuaternion();
    transX_ = 0;
    transY_ = 0;
    zoom_ = 0;
    globalScale_ = 1;
    position_ = QVector3D(0,0,1);
    center_ = QVector3D();
    theta_ = -90;
    phi_ = 0;
}
