#include "Geometry.h"

/** Compute intersection vertex between a ray and a triangle. Returne false if it doesn't intersect. */
bool Geometry::intersectionRayTriangle(const QVector3D &s1, const QVector3D &s2,
                                       const QVector3D &v1, const QVector3D &v2,
                                       const QVector3D &v3, const QVector3D &normal,
                                       QVector3D &vertInter)
{
    float dist1 = QVector3D::dotProduct(s1-v1, normal);
    float dist2 = QVector3D::dotProduct(s2-v1, normal);
    if ((dist1*dist2)>= 0.0f)
        return false;
    if (dist1==dist2)        //ray parallel to triangle plane
        return false;
    //intersection between ray and triangle
    vertInter = s1+(s2-s1)*(-dist1/(dist2-dist1));
    if (QVector3D::dotProduct(QVector3D::crossProduct(normal, v2-v1), vertInter-v1)<0.0f)
        return false;
    if (QVector3D::dotProduct(QVector3D::crossProduct(normal, v3-v2), vertInter-v2)<0.0f)
        return false;
    if (QVector3D::dotProduct(QVector3D::crossProduct(normal, v1-v3), vertInter-v1)<0.0f)
        return false;
    return true;
}

/** Intersection point between a ray and a plane */
QVector3D Geometry::intersectionRayPlane(const QVector3D &s1,const QVector3D &s2, const QVector3D &v1, const QVector3D &normal)
{
    float dist1 = QVector3D::dotProduct(s1-v1, normal);
    float dist2 = QVector3D::dotProduct(s2-v1, normal);
    if ((dist1*dist2)>= 0.0f)
        return QVector3D();
    if (dist1==dist2)        //ray parallel to triangle plane
        return QVector3D();
    //intersection between ray and triangle
    return s1+(s2-s1)*(-dist1/(dist2-dist1));
}

/** Projection of mouse coordinate on sphere unit */
QVector3D Geometry::mouseOnUnitSphere(const QPointF &mouseXY)
{
    float tempZ = 1-mouseXY.x()*mouseXY.x()-mouseXY.y()*mouseXY.y();
    float mouseZ= tempZ>0 ? sqrtf(tempZ) : 0;
    QVector3D sourisSphere(mouseXY.x(), mouseXY.y(), mouseZ);
    return sourisSphere.normalized();
}

/** Normalize coordinate mouse between 0 and 1
 */
QPointF Geometry::normalizedMouse(const QPointF &mouseXY, const QSizeF viewport)
{
    return QPointF(2.0*mouseXY.x()/viewport.width()-1.0,1.0-(2.0*mouseXY.y()/viewport.height()));
}

/** Compute the ray normal to screen (through mouse) */
void Geometry::getRayNormalToScreen(const QPointF &mouseXY, const QSizeF &viewport, QVector3D &vertexNear, QVector3D &vertexFar)
{
    double model[16];
    double projection[16];
    int view[4];
    double dX, dY, dZ;
    glGetIntegerv(GL_VIEWPORT, view);
    glGetDoublev (GL_MODELVIEW_MATRIX, model);
    glGetDoublev (GL_PROJECTION_MATRIX, projection);
    gluUnProject (mouseXY.x(), viewport.height()-mouseXY.y(),0.0, model, projection, view, &dX, &dY, &dZ);
    vertexNear.setX((float)dX);
    vertexNear.setY((float)dY);
    vertexNear.setZ((float)dZ);
    gluUnProject (mouseXY.x(), viewport.height()-mouseXY.y(),1.0, model, projection, view, &dX, &dY, &dZ);
    vertexFar.setX((float)dX);
    vertexFar.setY((float)dY);
    vertexFar.setZ((float)dZ);
}

/** Compute the unit vector coplanar to screen */
QVector3D Geometry::getCoplanarVectorOnScreen(const QPointF &mouseXYOld, const QPointF &mouseXYNew, const QSizeF &viewport)
{
    double model[16];
    double projection[16];
    int view[4];
    double dX1, dX2, dY1, dY2, dZ1, dZ2;
    glGetIntegerv(GL_VIEWPORT, view);
    glGetDoublev (GL_MODELVIEW_MATRIX, model);
    glGetDoublev (GL_PROJECTION_MATRIX, projection);
    gluUnProject (mouseXYOld.x(), viewport.height()-mouseXYOld.y(), 0.0, model, projection, view, &dX1, &dY1, &dZ1);
    gluUnProject (mouseXYNew.x(), viewport.height()-mouseXYNew.y(), 0.0, model, projection, view, &dX2, &dY2, &dZ2);
    return QVector3D(dX2 - dX1, dY2 - dY1, dZ2 -dZ1).normalized();
}

/** Compute the projection of a vertex on a line */
QVector3D Geometry::vertexOnLine(const QVector3D &vertex, const QVector3D &vertexNear, const QVector3D &vertexFar)
{
    QVector3D ab = vertexFar - vertexNear;
    float abSquared = ab.lengthSquared();
    float dot = QVector3D::dotProduct(ab, (vertex - vertexNear));
    float t = dot / abSquared;
    return (vertexNear + ab*t);
}

/** Project the mouse coordinate into the world coordinate at a given z */
QVector3D Geometry::point2Dto3D(const QPointF &mouseXY, const QSizeF &viewport, float z)
{
    double model[16];
    double projection[16];
    int view[4];
    double dX, dY, dZ;
    glGetIntegerv(GL_VIEWPORT, view);
    glGetDoublev (GL_MODELVIEW_MATRIX, model);
    glGetDoublev (GL_PROJECTION_MATRIX, projection);
    gluUnProject (mouseXY.x(), viewport.height()-mouseXY.y(), z, model, projection, view, &dX, &dY, &dZ);
    return QVector3D(dX, dY, dZ);
}

/** Project a vertex onto the screen */
QPointF Geometry::point3Dto2D(const QVector3D &vertex, const QSizeF &viewport, double &z)
{
    double model[16];
    double projection[16];
    int view[4];
    double dX, dY;
    glGetIntegerv(GL_VIEWPORT, view);
    glGetDoublev (GL_MODELVIEW_MATRIX, model);
    glGetDoublev (GL_PROJECTION_MATRIX, projection);
    gluProject(vertex.x(), vertex.y(), vertex.z(), model, projection, view, &dX, &dY, &z);
    return QPointF(dX, viewport.height()-dY);
}

/** Return any perpendicular vector to another vector */
QVector3D Geometry::getPerpendicularVector(const QVector3D &vec)
{
    QVector3D result;
    if (vec.x() == 0.0f || vec.y() == 0.0f || vec.z() == 0.0f)
    {
        if (vec.x() == 0.0f)
            result.setX(1.0f);
        else if (vec.y() == 0.0f)
            result.setY(1.0f);
        else
            result.setZ(1.0f);
    }
    else
    {
        result.setX(vec.z());
        result.setY(vec.z());
        result.setZ(-(vec.x()+vec.y()));
        result.normalize();
    }
    return result;
}

/** Compute the bounding box of a triangle defined by three vertices */
Aabb Geometry::computeTriangleAabb(const QVector3D &v1, const QVector3D &v2, const QVector3D &v3)
{
    float minX = std::min(std::min(v1.x(),v2.x()),v3.x());
    float minY = std::min(std::min(v1.y(),v2.y()),v3.y());
    float minZ = std::min(std::min(v1.z(),v2.z()),v3.z());

    float maxX = std::max(std::max(v1.x(),v2.x()),v3.x());
    float maxY = std::max(std::max(v1.y(),v2.y()),v3.y());
    float maxZ = std::max(std::max(v1.z(),v2.z()),v3.z());

    return Aabb(QVector3D(minX, minY, minZ), QVector3D(maxX, maxY, maxZ));
}

/** Compute the inradius (radius of incircle) of a triangle */
float Geometry::computeInradiusSquared(const QVector3D &v1, const QVector3D &v2, const QVector3D &v3)
{
    float a = (v1-v2).length();
    float b = (v1-v3).length();
    float c = (v2-v3).length();
    float halfPerimeter = (a+b+c)*0.5f;
    return (1/halfPerimeter)*(halfPerimeter-a)*(halfPerimeter-b)*(halfPerimeter-c);
}

/** If point is inside the triangle, test the sum of the areas */
bool Geometry::pointInsideTriangle(const QVector3D &point, const QVector3D &v1, const QVector3D &v2, const QVector3D &v3)
{
    QVector3D vec1 = v1-v2;
    QVector3D vec2 = v1-v3;
    QVector3D vecP1 = point-v2;
    QVector3D vecP2 = point-v3;
    float total = QVector3D::crossProduct(vec1, vec2).length();
    float area1 = QVector3D::crossProduct(vec1, vecP1).length();
    float area2 = QVector3D::crossProduct(vec2, vecP2).length();
    float area3 = QVector3D::crossProduct(vecP1, vecP2).length();
    if(fabs(total-(area1+area2+area3))<0.0001f) //magic epsilon...
        return true;
    else
        return false;
}

/** If a sphere intersect a triangle */
bool Geometry::sphereIntersectTriangle(const QVector3D &point, float radiusSq,
                                       const QVector3D &v1, const QVector3D &v2, const QVector3D &v3)
{
    if(Geometry::distanceToSegment(point,v1,v2)<radiusSq) return true;
    if(Geometry::distanceToSegment(point,v2,v3)<radiusSq) return true;
    if(Geometry::distanceToSegment(point,v1,v3)<radiusSq) return true;
    return false;
}

/** Minimum distance to a segment */
float Geometry::distanceToSegment(const QVector3D &point, const QVector3D &v1, const QVector3D &v2)
{
    float length = (v2-v1).lengthSquared();
    float t = QVector3D::dotProduct(point-v1,v2-v1)/length;
    if(t < 0.0) return (point-v1).lengthSquared();
    if(t> 1.0) return (point-v2).lengthSquared();
    QVector3D ptProj = v1 + t*(v2-v1);
    return (point-ptProj).lengthSquared();
}
