#include "Widget3D.h"

/** Constructor */
Widget3D::Widget3D(Window *window) : QGLWidget(QGLFormat(QGL::SampleBuffers), window), mesh_(0), camera_(), picking_(),
    sculpt_(), mouseXY_(), lastMouseXY_(), window_(window), cameraTimer_(0), fpsTime_(), fpsCounter_(0), symmetry_(false)
{
    setFocusPolicy(Qt::StrongFocus);
    setMouseTracking(true);
    startTimer(16);
}

/** Destructor */
Widget3D::~Widget3D()
{
    if(mesh_)
        delete mesh_;
}

/** OpenGL window properties */
QSize Widget3D::minimumSizeHint() const { return QSize(200, 200); }
QSize Widget3D::sizeHint() const { return QSize(QApplication::desktop()->width(), QApplication::desktop()->height()); }
void Widget3D::resizeGL(int width, int height) { camera_.setViewport(QSizeF(width, height)); camera_.updateProjection();}

/** Getters/Setters */
void Widget3D::setSculptRadius(int radius) { picking_.setRadiusScreen(std::max(20,radius)); }
void Widget3D::setSculptIntensity(int intensity) { sculpt_.setIntensity(intensity/200.f); }
void Widget3D::setSculptMode(int mode) { sculpt_.setSculptMode(Sculpt::SculptMode(mode)); }
void Widget3D::setTopoMode(int mode) { sculpt_.setTopoMode(Sculpt::TopoMode(mode)); }
void Widget3D::setTopoDetail(int detail) { sculpt_.setDetail(detail/200.f); }
void Widget3D::setSculptCulling() { sculpt_.toggleCulling(); }
void Widget3D::setSymmetry() { symmetry_ = !symmetry_; }
void Widget3D::setCamera(int mode) { camera_.setMode(Camera::CameraMode(mode)); }

/** Initialize openGL */
void Widget3D::initializeGL()
{
    glEnable(GL_DEPTH_TEST);
    glClearColor(0.2, 0.2, 0.2, 0);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

/** Draw on screen */
void Widget3D::paintGL()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    camera_.updateView();
    camera_.loadProjectionMatrix(); //for immediate mode...
    camera_.loadViewMatrix();       //for immediate mode...
    if(mesh_)
        mesh_->draw(camera_, picking_.getIntersectionPoint(), picking_.getRadiusWorldSquared());
    showFps();
}

/** Key press event */
void Widget3D::keyPressEvent(QKeyEvent *event)
{
    if(event->isAutoRepeat())
        return;
    if(event->key()==Qt::Key_O && mesh_)
        mesh_->toggleDisplayOctree();
    if (event->key()==Qt::Key_Q || event->key()==Qt::Key_Left)
        camera_.setMoveX(-1);
    else if (event->key()==Qt::Key_D || event->key()==Qt::Key_Right)
        camera_.setMoveX(1);
    if (event->key()==Qt::Key_Z || event->key()==Qt::Key_Up)
        camera_.setMoveZ(-1);
    else if (event->key()==Qt::Key_S || event->key()==Qt::Key_Down)
        camera_.setMoveZ(1);
    if(!cameraTimer_)
        cameraTimer_ = startTimer(20);
}

/** Key release event */
void Widget3D::keyReleaseEvent(QKeyEvent *event)
{
    if(event->isAutoRepeat())
        return;
    if (event->key()==Qt::Key_Q || event->key()==Qt::Key_D || event->key()==Qt::Key_Left || event->key()==Qt::Key_Right)
        camera_.setMoveX(0);
    if (event->key()==Qt::Key_Z || event->key()==Qt::Key_S || event->key()==Qt::Key_Up || event->key()==Qt::Key_Down)
        camera_.setMoveZ(0);
    if(cameraTimer_ && !camera_.isMoving())
    {
        killTimer(cameraTimer_);
        cameraTimer_=0;
    }
}

/** Mouse press event */
void Widget3D::mousePressEvent(QMouseEvent *event)
{
    lastMouseXY_ = event->posF();
    if (event->buttons() == Qt::LeftButton)
        camera_.start(event->posF());
    else if (event->buttons() == Qt::RightButton)
    {
        if(mesh_)
        {
            mesh_->startPushState();
            if(picking_.getMesh())
                sculpt_.setSweepCenter(picking_.getIntersectionPoint());
        }
    }
}

/** Mouse move event */
void Widget3D::mouseMoveEvent(QMouseEvent *event)
{
    QVector3D vertexNear,vertexFar;
    Geometry::getRayNormalToScreen(event->posF(), size(), vertexNear, vertexFar);
    if(mesh_)
    {
        vertexNear = mesh_->getInverseTransformation()*vertexNear;
        vertexFar = mesh_->getInverseTransformation()*vertexFar;
        picking_.intersectionRayMesh(mesh_, vertexNear, vertexFar, event->posF(), size());
    }
    if (event->buttons() == Qt::LeftButton)
        camera_.rotate(event->posF());
    else if (event->buttons() == Qt::MidButton)
        camera_.translate((event->posF()-mouseXY_)/3000.f);
    else if (event->buttons() == Qt::RightButton)
    {
        if ((picking_.getMesh() || sculpt_.isSweep()) && mesh_ &&
                QLineF(event->posF(),lastMouseXY_).length()>0.1f*picking_.getRadiusScreen())
        {
            if(sculpt_.isSweep())
            {
                sculpt_.setSweepCenter(Geometry::vertexOnLine(sculpt_.getSweepCenter(),vertexNear, vertexFar));
                sculpt_.setSweepDir(Geometry::getCoplanarVectorOnScreen(lastMouseXY_, event->posF(), size()));
                picking_.forceCenter(mesh_, sculpt_.getSweepCenter(), event->posF(), size());
            }
            float radiusWorldSquared = picking_.getRadiusWorldSquared();
            QVector3D intersectionPoint = picking_.getIntersectionPoint();
            picking_.setPickedVerticesInsideSphere(radiusWorldSquared);
            sculpt_.setMesh(picking_.getMesh());
            sculpt_.sculptMesh(picking_.getSelectedVertices(), intersectionPoint,
                               radiusWorldSquared , (vertexFar-vertexNear).normalized());

            if(symmetry_ && !sculpt_.isSweep())
            {
                float dist = vertexNear.distanceToPlane(QVector3D(), QVector3D(1,0,0));
                vertexNear -= QVector3D(1,0,0)*dist*2.f;

                dist = vertexFar.distanceToPlane(QVector3D(), QVector3D(1,0,0));
                vertexFar -= QVector3D(1,0,0)*dist*2.f;

                QVector3D symmetricIntersection = picking_.getIntersectionPoint();
                dist = symmetricIntersection.distanceToPlane(QVector3D(), QVector3D(1,0,0));
                symmetricIntersection  -= QVector3D(1,0,0)*dist*2.f;

                picking_.intersectionRayMesh(mesh_, vertexNear, vertexFar, event->posF(), size());
                picking_.setMesh(mesh_);
                picking_.setIntersectionPoint(symmetricIntersection);
                picking_.setPickedVerticesInsideSphere(radiusWorldSquared);
                sculpt_.sculptMesh(picking_.getSelectedVertices(), picking_.getIntersectionPoint(),
                                   radiusWorldSquared , (vertexFar-vertexNear).normalized());
                picking_.setIntersectionPoint(intersectionPoint);
                picking_.setRadiusWorldSquared(radiusWorldSquared);
            }

            updateMeshInformation();
            lastMouseXY_ = event->posF();
        }
    }
    mouseXY_ = event->posF();
    updateGL();
}

/** Mouse release event */
void Widget3D::mouseReleaseEvent(QMouseEvent *event)
{
    if(mesh_)
        mesh_->checkLeavesUpdate();
}

/** Wheel event */
void Widget3D::wheelEvent(QWheelEvent *event)
{
    if (event->orientation() == Qt::Vertical)
        camera_.zoom(event->delta()/10000.f);
}

/** Timer event */
void Widget3D::timerEvent(QTimerEvent *event)
{
    if(event->timerId()==cameraTimer_)
        camera_.updateTranslation();
    updateGL();
}

/** Set mesh Color */
void Widget3D::setMeshColor()
{
    if (mesh_)
    {
        QColor color = QColorDialog::getColor(mesh_->getColor(), this, "mesh color");
        if (color.isValid())
        {
            mesh_->setColor(color);
            if (QPushButton *colorButton = window_->getColorButton())
                colorButton->setPalette(QPalette(QColor(color)));
        }
    }
}

/** Set mesh shader */
void Widget3D::setMeshShader(int mode)
{
    if (mesh_)
        mesh_->setShader(Mesh::ShaderMode(mode));
}

/** Load file */
void Widget3D::loadFile()
{
    Files files;
    QString fileName = QFileDialog::getOpenFileName(this, tr("load file"), "/home", "Files(*.stl *.obj *.3ds)");
    if (!fileName.isNull())
    {
        //reset flags... not necessary
        Mesh::stateMask_= 1;
        Vertex::tagMask_ = 1;
        Vertex::sculptMask_ = 1;
        Triangle::tagMask_ = 1;

        Mesh *mesh = 0;
        if(fileName.endsWith(".stl",Qt::CaseInsensitive))
            mesh = files.loadSTL(fileName.toStdString(), window_);
        else if(fileName.endsWith(".obj",Qt::CaseInsensitive))
            mesh = files.loadOBJ(fileName.toStdString(), window_);
        else if(fileName.endsWith(".3ds",Qt::CaseInsensitive))
            mesh = files.load3DS(fileName.toStdString(), window_);
        if(mesh_)
            delete mesh_;
        mesh_ = mesh;
        camera_.reset();
        float length = (mesh_->getOctree()->getAabbLoose().max_-mesh_->getOctree()->getAabbLoose().min_).length();
        camera_.setGlobalScale(length);
        camera_.zoom(-0.4f);
        if (QPushButton *colorButton = window_->getColorButton())
            colorButton->setPalette(QPalette(QColor(mesh_->getColor())));
        if (QButtonGroup *shaderButtonGroup = window_->getShaderButtonGroup())
            shaderButtonGroup->button(mesh_->getShader())->setChecked(true);
        updateMeshInformation();
    }
}

/** Save file */
void Widget3D::saveFile()
{
    Files files;
    QString fileName = QFileDialog::getSaveFileName(this, tr("save file"), "/home", tr("STL Files (*.stl )"));
    if (!fileName.isNull() && mesh_)
        files.saveSTL(mesh_, fileName.toStdString(), window_);
}

/** Show frame per second */
void Widget3D::showFps()
{
    ++fpsCounter_;
    int fps = 1000.f*(float)fpsCounter_/fpsTime_.elapsed();
    QString stringFps=QString("FPS : ").append(QString::number(fps));
    window_->statusBar()->showMessage(window_->statusBar()->currentMessage().leftJustified(60,' ',true).append(stringFps));
    if(fpsTime_.elapsed()>1000)
    {
        fpsTime_.restart();
        fpsCounter_=0;
    }
}

/** Update mesh information on the status bar */
void Widget3D::updateMeshInformation()
{
    if(mesh_)
    {
        QString meshInfo = QString(" Triangles : ").append(QString::number(mesh_->getNbTriangles())).
                append(QString("     Vertices : ")).append(QString::number(mesh_->getNbVertices()));
        window_->statusBar()->showMessage(meshInfo);
    }
}

/** Undo */
void Widget3D::undo()
{
    if(mesh_)
    {
        mesh_->undo();
        updateMeshInformation();
    }
}

/** Redo */
void Widget3D::redo()
{
    if(mesh_)
    {
        mesh_->redo();
        updateMeshInformation();
    }
}
