#include "Aabb.h"

/** Constructor  */
Aabb::Aabb(const QVector3D &min, const QVector3D &max) : min_(min), max_(max)
{}

/** Destructor  */
Aabb::~Aabb()
{}

/** Setters/Getters */
QVector3D Aabb::getCenter() const { return (max_+min_)/2; }

/** Collision detection) */
bool Aabb::isOutside(const Aabb &aabb) const
{
    if( aabb.min_.x() > max_.x() || aabb.max_.x() < min_.x() )
        return true;
    if( aabb.min_.y() > max_.y() || aabb.max_.y() < min_.y() )
        return true;
    if( aabb.min_.z() > max_.z() || aabb.max_.z() < min_.z() )
        return true;
    return false;
}

/** Return true if aabb is inside the box */
bool Aabb::isInside(const Aabb &aabb)const
{
    if( min_.x() >= aabb.min_.x() && max_.x() <= aabb.max_.x() &&
            min_.y() >= aabb.min_.y() && max_.y() <= aabb.max_.y() &&
            min_.z() >= aabb.min_.z() && max_.z() <= aabb.max_.z() )
        return true;
    return false;
}

/** Return true if vert is inside the aabb */
bool Aabb::pointInside(const QVector3D &vert) const
{
    if(vert.x()<=min_.x())
        return false;
    if(vert.x()>max_.x())
        return false;
    if(vert.y()<=min_.y())
        return false;
    if(vert.y()>max_.y())
        return false;
    if(vert.z()<=min_.z())
        return false;
    if(vert.z()>max_.z())
        return false;
    return true;
}

/** Change the size of the aabb to include vert */
void Aabb::expand(const QVector3D &vert)
{
    if (vert.x()>max_.x())
        max_.setX(vert.x());
    if (vert.x()<min_.x())
        min_.setX(vert.x());
    if (vert.y()>max_.y())
        max_.setY(vert.y());
    if (vert.y()<min_.y())
        min_.setY(vert.y());
    if (vert.z()>max_.z())
        max_.setZ(vert.z());
    if (vert.z()<min_.z())
        min_.setZ(vert.z());
}

/** Change the size of the aabb to include another aabb */
void Aabb::expand(const Aabb &aabb)
{
    if (aabb.max_.x()>max_.x())
        max_.setX(aabb.max_.x());
    if (aabb.min_.x()<min_.x())
        min_.setX(aabb.min_.x());
    if (aabb.max_.y()>max_.y())
        max_.setY(aabb.max_.y());
    if (aabb.min_.y()<min_.y())
        min_.setY(aabb.min_.y());
    if (aabb.max_.z()>max_.z())
        max_.setZ(aabb.max_.z());
    if (aabb.min_.z()<min_.z())
        min_.setZ(aabb.min_.z());
}

/** Return true if a ray intersection the box */
bool Aabb::intersectRay(const QVector3D &vert, const QVector3D &dir) const
{
    float t1 = (min_.x() - vert.x())/dir.x();
    float t2 = (max_.x() - vert.x())/dir.x();
    float t3 = (min_.y() - vert.y())/dir.y();
    float t4 = (max_.y() - vert.y())/dir.y();
    float t5 = (min_.z() - vert.z())/dir.z();
    float t6 = (max_.z() - vert.z())/dir.z();

    float tmin = std::max(std::max(std::min(t1, t2), std::min(t3, t4)), std::min(t5, t6));
    float tmax = std::min(std::min(std::max(t1, t2), std::max(t3, t4)), std::max(t5, t6));
    return (tmax >=0 && tmin<tmax);
}

/** Return true if a sphere intersect with the box */
bool Aabb::intersectSphere(const QVector3D &vert, float radiusSquared) const
{
    QVector3D nearest;
    if (min_.x() > vert.x())
        nearest.setX(min_.x());
    else if (max_.x() < vert.x())
        nearest.setX(max_.x());
    else
        nearest.setX(vert.x());
    if (min_.y() > vert.y())
        nearest.setY(min_.y());
    else if (max_.y() < vert.y())
        nearest.setY(max_.y());
    else
        nearest.setY(vert.y());
    if (min_.z() > vert.z())
        nearest.setZ(min_.z());
    else if (max_.z() < vert.z())
        nearest.setZ(max_.z());
    else
        nearest.setZ(vert.z());
    return (vert-nearest).lengthSquared()<radiusSquared;
}

/** Check if the aabb is a plane */
void Aabb::checkFlat(float offset)
{
    if(min_.x()==max_.x())
    {
        min_.setX(min_.x()-offset);
        max_.setX(max_.x()+offset);
    }
    if(min_.y()==max_.y())
    {
        min_.setY(min_.y()-offset);
        max_.setY(max_.y()+offset);
    }
    if(min_.z()==max_.z())
    {
        min_.setZ(min_.z()-offset);
        max_.setZ(max_.z()+offset);
    }
}

/** Draw the aabb */
void Aabb::draw(const QColor &color) const
{
    glBegin(GL_LINES);
    glColor3f(color.redF(), color.greenF(), color.blueF());
    glVertex3f(min_.x(), min_.y(), min_.z());glVertex3f(max_.x(), min_.y(), min_.z());
    glVertex3f(max_.x(), min_.y(), min_.z());glVertex3f(max_.x(), max_.y(), min_.z());
    glVertex3f(max_.x(), max_.y(), min_.z());glVertex3f(min_.x(), max_.y(), min_.z());
    glVertex3f(min_.x(), max_.y(), min_.z());glVertex3f(min_.x(), min_.y(), min_.z());
    glVertex3f(min_.x(), min_.y(), min_.z());glVertex3f(min_.x(), min_.y(), max_.z());
    glVertex3f(min_.x(), min_.y(), max_.z());glVertex3f(max_.x(), min_.y(), max_.z());
    glVertex3f(max_.x(), min_.y(), max_.z());glVertex3f(max_.x(), min_.y(), min_.z());
    glVertex3f(max_.x(), min_.y(), max_.z());glVertex3f(max_.x(), max_.y(), max_.z());
    glVertex3f(max_.x(), max_.y(), max_.z());glVertex3f(max_.x(), max_.y(), min_.z());
    glVertex3f(max_.x(), max_.y(), max_.z());glVertex3f(min_.x(), max_.y(), max_.z());
    glVertex3f(min_.x(), max_.y(), max_.z());glVertex3f(min_.x(), max_.y(), min_.z());
    glVertex3f(min_.x(), max_.y(), max_.z());glVertex3f(min_.x(), min_.y(), max_.z());
    glEnd();
}
