#include "Mesh.h"
#include <iostream>

int Mesh::stateMask_ = 1;

/** Constructor */
Mesh::Mesh() : vertices_(), triangles_(), verticesBuffer_(QGLBuffer::VertexBuffer), normalsBuffer_(QGLBuffer::VertexBuffer),
    indicesBuffer_(QGLBuffer::IndexBuffer), center_(), shaderType_(PHONG), shaderProgram_(), color_(120, 200, 200), scale_(1),
    octree_(0), displayOctree_(false), matTransform_(), matTransformArray_(), reflectionTexture_(), leavesUpdate_(), undo_(),
    redo_(), undoIte_(), beginIte_(false)
{
    updateTransformation();
}

/** Destructor */
Mesh::~Mesh()
{
    delete octree_;
}

/** Setters/Getters */
std::vector<Triangle>& Mesh::getTriangles() { return triangles_; }
std::vector<Vertex>& Mesh::getVertices() { return vertices_; }
std::vector<Octree*>& Mesh::getLeavesUpdate() { return leavesUpdate_; }
Triangle& Mesh::getTriangle(int i) { return triangles_[i]; }
Vertex& Mesh::getVertex(int i) { return vertices_[i]; }
int Mesh::getNbTriangles() const { return triangles_.size(); }
int Mesh::getNbVertices() const { return vertices_.size(); }
QVector3D Mesh::getCenter() const { return center_; }
Octree* Mesh::getOctree() const { return octree_; }
QColor Mesh::getColor() const { return color_; }
float Mesh::getScale() const { return scale_; }
void Mesh::setColor(const QColor &color) { color_=color; }
bool Mesh::getDisplayOctree() const { return displayOctree_; }
void Mesh::toggleDisplayOctree() { displayOctree_=!displayOctree_; }
QMatrix4x4& Mesh::getTransformation() { return matTransform_; }
QMatrix4x4 Mesh::getInverseTransformation() const { return matTransform_.inverted(); }
Mesh::ShaderMode Mesh::getShader() const { return shaderType_; }

/** Return all the triangles linked to a group of vertices */
std::vector<int> Mesh::getTrianglesFromVertices(const std::vector<int> &iVerts)
{
    ++Triangle::tagMask_;
    std::vector<int> triangles;
    int nbVerts = iVerts.size();
    for(int i=0;i<nbVerts;++i)
    {
        std::vector<int> &iTris = vertices_[iVerts[i]].tIndices_;
        int nbTris = iTris.size();
        for(int j=0;j<nbTris;++j)
        {
            int iTri = iTris[j];
            if(triangles_[iTri].tagFlag_!=Triangle::tagMask_)
            {
                triangles.push_back(iTri);
                triangles_[iTri].tagFlag_=Triangle::tagMask_;
            }
        }
    }
    return triangles;
}

/** Return all the triangles linked to a group of vertices */
std::vector<int> Mesh::getVerticesFromTriangles(const std::vector<int> &iTris)
{
    ++Vertex::tagMask_;
    std::vector<int> vertices;
    int nbTris = iTris.size();
    for(int i=0;i<nbTris;++i)
    {
        Triangle &t=triangles_[iTris[i]];
        int iVer1 = t.vIndices_[0];
        int iVer2 = t.vIndices_[1];
        int iVer3 = t.vIndices_[2];
        if(vertices_[iVer1].tagFlag_!=Vertex::tagMask_)
        {
            vertices.push_back(iVer1);
            vertices_[iVer1].tagFlag_=Vertex::tagMask_;
        }
        if(vertices_[iVer2].tagFlag_!=Vertex::tagMask_)
        {
            vertices.push_back(iVer2);
            vertices_[iVer2].tagFlag_=Vertex::tagMask_;
        }
        if(vertices_[iVer3].tagFlag_!=Vertex::tagMask_)
        {
            vertices.push_back(iVer3);
            vertices_[iVer3].tagFlag_=Vertex::tagMask_;
        }
    }
    return vertices;
}

/** Get more triangles (n-ring) */
void Mesh::expandTriangles(std::vector<int> &iTris, int nRing)
{
    ++Triangle::tagMask_;
    int nbTris = iTris.size();
    for(int i=0;i<nbTris;++i)
        triangles_[iTris[i]].tagFlag_ = Triangle::tagMask_;
    int iBegin = 0;
    while(nRing)
    {
        --nRing;
        for(int i=iBegin;i<nbTris;++i)
        {
            Triangle &t=triangles_[iTris[i]];
            std::vector<int> &iTris1 = vertices_[t.vIndices_[0]].tIndices_;
            std::vector<int> &iTris2 = vertices_[t.vIndices_[1]].tIndices_;
            std::vector<int> &iTris3 = vertices_[t.vIndices_[2]].tIndices_;
            int nbTris1 = iTris1.size();
            int nbTris2 = iTris2.size();
            int nbTris3 = iTris3.size();
            for(int j=0;j<nbTris1;++j)
            {
                Triangle &t = triangles_[iTris1[j]];
                if(t.tagFlag_!=Triangle::tagMask_)
                {
                    t.tagFlag_ = Triangle::tagMask_;
                    iTris.push_back(iTris1[j]);
                }
            }
            for(int j=0;j<nbTris2;++j)
            {
                Triangle &t = triangles_[iTris2[j]];
                if(t.tagFlag_!=Triangle::tagMask_)
                {
                    t.tagFlag_ = Triangle::tagMask_;
                    iTris.push_back(iTris2[j]);
                }
            }
            for(int j=0;j<nbTris3;++j)
            {
                Triangle &t = triangles_[iTris3[j]];
                if(t.tagFlag_!=Triangle::tagMask_)
                {
                    t.tagFlag_ = Triangle::tagMask_;
                    iTris.push_back(iTris3[j]);
                }
            }
        }
        iBegin = nbTris;
        nbTris = iTris.size();
    }
}

/** Get more vertices (n-ring) */
void Mesh::expandVertices(std::vector<int> &iVerts, int nRing)
{
    ++Vertex::tagMask_;
    int nbVerts = iVerts.size();
    for(int i=0;i<nbVerts;++i)
        vertices_[iVerts[i]].tagFlag_ = Vertex::tagMask_;
    int iBegin = 0;
    while(nRing)
    {
        --nRing;
        for(int i=iBegin;i<nbVerts;++i)
        {
            std::vector<int> &ring = vertices_[iVerts[i]].ringVertices_;
            int nbRing = ring.size();
            for(int j=0;j<nbRing;++j)
            {
                Vertex &vRing = vertices_[ring[j]];
                if(vRing.tagFlag_!=Vertex::tagMask_)
                {
                    vRing.tagFlag_ = Vertex::tagMask_;
                    iVerts.push_back(ring[j]);
                }
            }
        }
        iBegin = nbVerts;
        nbVerts = iVerts.size();
    }
}

/** Compute the vertices around a vertex */
void Mesh::computeRingVertices(int iVert)
{
    ++Vertex::tagMask_;
    std::vector<int> &iTris = vertices_[iVert].tIndices_;
    std::vector<int> &ring = vertices_[iVert].ringVertices_;
    ring.clear();
    int nbTris = iTris.size();
    for(int i=0;i<nbTris;++i)
    {
        const Triangle &t=triangles_[iTris[i]];
        int iVer1 = t.vIndices_[0];
        int iVer2 = t.vIndices_[1];
        int iVer3 = t.vIndices_[2];
        if(iVer1!=iVert && vertices_[iVer1].tagFlag_!=Vertex::tagMask_)
        {
            ring.push_back(iVer1);
            vertices_[iVer1].tagFlag_=Vertex::tagMask_;
        }
        if(iVer2!=iVert && vertices_[iVer2].tagFlag_!=Vertex::tagMask_)
        {
            ring.push_back(iVer2);
            vertices_[iVer2].tagFlag_=Vertex::tagMask_;
        }
        if(iVer3!=iVert && vertices_[iVer3].tagFlag_!=Vertex::tagMask_)
        {
            ring.push_back(iVer3);
            vertices_[iVer3].tagFlag_=Vertex::tagMask_;
        }
    }
}

/** Set shader on the mesh */
void Mesh::setShader(ShaderMode shader)
{
    shaderType_ = shader;
    shaderProgram_.removeAllShaders();
    QString pathShader = QDir::currentPath()+"/Shaders/";
    switch(shaderType_)
    {
    case PHONG :
        shaderProgram_.addShaderFromSourceFile(QGLShader::Vertex, pathShader+"phongVertex.glsl");
        shaderProgram_.addShaderFromSourceFile(QGLShader::Fragment, pathShader+"phongFragment.glsl");
        break;
    case WIREFRAME :
        shaderProgram_.addShaderFromSourceFile(QGLShader::Vertex, pathShader+"wireframeVertex.glsl");
        shaderProgram_.addShaderFromSourceFile(QGLShader::Geometry, pathShader+"wireframeGeometry.glsl");
        shaderProgram_.addShaderFromSourceFile(QGLShader::Fragment, pathShader+"wireframeFragment.glsl");
        shaderProgram_.setGeometryInputType(GL_TRIANGLES);
        shaderProgram_.setGeometryOutputType(GL_TRIANGLE_STRIP);
        shaderProgram_.setGeometryOutputVertexCount(3);
        break;
    case TRANSPARENCY :
        shaderProgram_.addShaderFromSourceFile(QGLShader::Vertex, pathShader+"transparencyVertex.glsl");
        shaderProgram_.addShaderFromSourceFile(QGLShader::Fragment, pathShader+"transparencyFragment.glsl");
        break;
    case REFLECTION_1 :
    case REFLECTION_2 :
        QImage tex;
        bool loaded=false;
        if (shaderType_==REFLECTION_1) loaded = tex.load(QDir::currentPath()+"/Ressources/reflection1.png");
        else if (shaderType_==REFLECTION_2) loaded = tex.load(QDir::currentPath()+"/Ressources/reflection2.png");
        if(loaded)
        {
            tex = QGLWidget::convertToGLFormat(tex);
            glGenTextures( 1, &reflectionTexture_ );
            glBindTexture( GL_TEXTURE_2D, reflectionTexture_ );
            glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,tex.width(),tex.height(),0,GL_RGBA,GL_UNSIGNED_BYTE,tex.bits());
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
        }
        shaderProgram_.addShaderFromSourceFile(QGLShader::Vertex, pathShader+"reflectionVertex.glsl");
        shaderProgram_.addShaderFromSourceFile(QGLShader::Fragment, pathShader+"reflectionFragment.glsl");
        break;
    }
    shaderProgram_.link();
}

/** Return center of a triangle */
QVector3D Mesh::getTriangleCenter(int iTri) const
{
    const Triangle &t=triangles_[iTri];
    return (vertices_[t.vIndices_[0]]+vertices_[t.vIndices_[1]]+vertices_[t.vIndices_[2]])/3;
}

/** Move the mesh center to a certain point */
void Mesh::moveTo(const QVector3D &destination)
{
    matTransform_.setToIdentity();
    matTransform_.translate(destination-center_);
    updateTransformation();
}

/** Set transformation */
void Mesh::setTransformation(QMatrix4x4 matTransform)
{
    matTransform_=matTransform;
    updateTransformation();
}

/** Update transformation */
void Mesh::updateTransformation()
{
    qreal *dataMat = matTransform_.data();
    for (int i= 0; i < 16; ++i)
        matTransformArray_[i] = dataMat[i];
}

/** Draw the mesh */
void Mesh::draw(const Camera &camera, const QVector3D &centerPicking, float radiusSquared)
{
    QMatrix4x4 modelViewMatrix = camera.getViewMatrix()*matTransform_;
    QMatrix4x4 modelViewProjectionMatrix = camera.getProjectionMatrix()*modelViewMatrix;
    switch (shaderType_)
    {
    case PHONG :
        shaderProgram_.bind();
        shaderProgram_.setUniformValue("modelViewMatrix", modelViewMatrix);
        shaderProgram_.setUniformValue("modelViewProjectionMatrix", modelViewProjectionMatrix);
        shaderProgram_.setUniformValue("normalMatrix", modelViewMatrix.normalMatrix());
        shaderProgram_.setUniformValue("centerPicking", modelViewMatrix*centerPicking);
        shaderProgram_.setUniformValue("radiusSquared", radiusSquared);
        shaderProgram_.setUniformValue("color", color_);
        drawBufferShader();
        shaderProgram_.release();
        break;
    case WIREFRAME :
        shaderProgram_.bind();
        shaderProgram_.setUniformValue("modelViewMatrix", modelViewMatrix);
        shaderProgram_.setUniformValue("modelViewProjectionMatrix", modelViewProjectionMatrix);
        shaderProgram_.setUniformValue("centerPicking", modelViewMatrix*centerPicking);
        shaderProgram_.setUniformValue("radiusSquared", radiusSquared);
        shaderProgram_.setUniformValue("color", color_);
        shaderProgram_.setUniformValue("winScale", camera.getViewport());
        drawBufferShader();
        shaderProgram_.release();
        break;
    case TRANSPARENCY :
        glDepthMask(GL_FALSE);
        glEnable(GL_BLEND);
        shaderProgram_.bind();
        shaderProgram_.setUniformValue("modelViewMatrix", modelViewMatrix);
        shaderProgram_.setUniformValue("modelViewProjectionMatrix", modelViewProjectionMatrix);
        shaderProgram_.setUniformValue("normalMatrix", modelViewMatrix.normalMatrix());
        shaderProgram_.setUniformValue("centerPicking", modelViewMatrix*centerPicking);
        shaderProgram_.setUniformValue("radiusSquared", radiusSquared);
        shaderProgram_.setUniformValue("lightPosition", center_);
        shaderProgram_.setUniformValue("color", QColor::fromRgb(color_.red(),color_.green(), color_.blue(), 35));
        drawBufferShader();
        shaderProgram_.release();
        glDisable(GL_BLEND);
        glDepthMask(GL_TRUE);
        break;
    case REFLECTION_1 :
    case REFLECTION_2 :
        shaderProgram_.bind();
        shaderProgram_.setUniformValue("modelViewMatrix", modelViewMatrix);
        shaderProgram_.setUniformValue("modelViewProjectionMatrix", modelViewProjectionMatrix);
        shaderProgram_.setUniformValue("normalMatrix", modelViewMatrix.normalMatrix());
        shaderProgram_.setUniformValue("centerPicking", modelViewMatrix*centerPicking);
        shaderProgram_.setUniformValue("radiusSquared", radiusSquared);
        shaderProgram_.setUniformValue("color", color_);
        shaderProgram_.setUniformValue("reflectionTexture", 0);
        glBindTexture(GL_TEXTURE_2D, reflectionTexture_);
        drawBufferShader();
        shaderProgram_.release();
        break;
    }
    if (displayOctree_)
    {
        glPushMatrix();
        glMultMatrixf(matTransformArray_);
        octree_->draw();
        glPopMatrix();
    }
}

/** Draw buffer shaders */
void Mesh::drawBufferShader()
{
    verticesBuffer_.bind();
    shaderProgram_.enableAttributeArray("vertex");
    shaderProgram_.setAttributeBuffer("vertex", GL_FLOAT,  0,  3);
    normalsBuffer_.bind();
    shaderProgram_.enableAttributeArray("normal");
    shaderProgram_.setAttributeBuffer("normal", GL_FLOAT,  0,  3);
    indicesBuffer_.bind();
    glDrawElements(GL_TRIANGLES, getNbTriangles()*3, GL_UNSIGNED_INT, 0);
    indicesBuffer_.release();
    shaderProgram_.disableAttributeArray("normal");
    shaderProgram_.disableAttributeArray("vertex");
    normalsBuffer_.release();
    verticesBuffer_.release();
}

/** Initialize Vertex Buffer Object (VBO) */
void Mesh::initVBO(Window *window)
{
    GLfloat* verticesArray=(GLfloat*)malloc(vertices_.size()*3*sizeof(GLfloat)*2);
    GLfloat* normalsArray=(GLfloat*)malloc(vertices_.size()*3*sizeof(GLfloat)*2);
    GLuint* indicesArray=(GLuint*)malloc(triangles_.size()*3*sizeof(GLuint)*2);
    if(window)
        window->statusBar()->showMessage("Loading : Init VBO 1");
    int nbVertices = getNbVertices();
    int nbTriangles = getNbTriangles();
#pragma omp parallel for
    for (int i=0;i<nbVertices;++i)
    {
        Vertex &ver=vertices_[i];
        verticesArray[i*3]=ver.x();
        verticesArray[i*3+1]=ver.y();
        verticesArray[i*3+2]=ver.z();
        const std::vector<int> &iTri=ver.tIndices_;
        int nbTri = iTri.size();
        Vertex normal;
        for (int j=0;j<nbTri;++j)
            normal+=triangles_[iTri[j]].normal_;
        normal.normalize();
        vertices_[i].normal_=normal;
        normalsArray[i*3]=normal.x();
        normalsArray[i*3+1]=normal.y();
        normalsArray[i*3+2]=normal.z();
    }
    if(window)
        window->statusBar()->showMessage("Loading : Init VBO 2");
#pragma omp parallel for
    for (int i=0;i<nbTriangles;++i)
    {
        Triangle &tri=triangles_[i];
        indicesArray[i*3]=tri.vIndices_[0];
        indicesArray[i*3+1]=tri.vIndices_[1];
        indicesArray[i*3+2]=tri.vIndices_[2];
    }
    verticesBuffer_.create();
    verticesBuffer_.setUsagePattern(QGLBuffer::DynamicDraw);
    verticesBuffer_.bind();
    verticesBuffer_.allocate(verticesArray, getNbVertices()*3*sizeof(GLfloat)*2);
    verticesBuffer_.release();
    normalsBuffer_.create();
    normalsBuffer_.setUsagePattern(QGLBuffer::DynamicDraw);
    normalsBuffer_.bind();
    normalsBuffer_.allocate(normalsArray, getNbVertices()*3*sizeof(GLfloat)*2);
    normalsBuffer_.release();
    indicesBuffer_.create();
    indicesBuffer_.setUsagePattern(QGLBuffer::DynamicDraw);
    indicesBuffer_.bind();
    indicesBuffer_.allocate(indicesArray, getNbTriangles()*3*sizeof(GLuint)*2);
    indicesBuffer_.release();
    free(verticesArray);
    free(normalsArray);
    free(indicesArray);
    setShader(shaderType_);
}

/** Initialize the mesh information : center, octree, scale ... */
void Mesh::initMesh(Window *window)
{
    int nbVertices = getNbVertices();
    int nbTriangles = getNbTriangles();
    Aabb aabb;
    aabb.min_ = vertices_[0];
    aabb.max_ = vertices_[0];
    if(window)
        window->statusBar()->showMessage("Loading : Init topology 1/2");
    for(int i=0;i<nbVertices;++i)
    {
        computeRingVertices(i);
        aabb.expand(vertices_[i]);
    }
    center_ = aabb.getCenter();
    float diag = (aabb.max_-aabb.min_).length();
    if(window)
        window->statusBar()->showMessage("Loading : scaling");
    scale_ = Mesh::globalScale_/diag;
#pragma omp parallel for
    for(int i=0;i<nbVertices;++i)
        vertices_[i]*=scale_;
    matTransform_.scale(scale_);
    aabb.max_*=scale_;
    aabb.min_*=scale_;
    center_*=scale_;
    if(window)
        window->statusBar()->showMessage("Loading : Init topology 2/2");
#pragma omp parallel for
    for(int i=0;i<nbTriangles;++i)
    {
        Triangle &t = triangles_[i];
        t.aabb_ = Geometry::computeTriangleAabb(vertices_[t.vIndices_[0]],vertices_[t.vIndices_[1]],vertices_[t.vIndices_[2]]);
    }
    if(window)
        window->statusBar()->showMessage("Loading : Init octree");
    
    aabb.checkFlat((aabb.max_-aabb.min_).length()*0.02);
    QVector3D vecShift = (aabb.max_-aabb.min_)*0.2; //root octree bigger than minimum aabb...
    aabb.min_-=vecShift;
    aabb.max_+=vecShift;
    std::vector<int> triangles(nbTriangles);
#pragma omp parallel for
    for (int i=0;i<nbTriangles;++i)
        triangles[i] = i;
    ++Triangle::tagMask_;
    if(octree_)
        delete octree_;
    octree_ = new Octree();
    octree_->build(this,triangles,aabb);
}

/** Update geometry  */
void Mesh::updateMesh(const std::vector<int> &iTris, const std::vector<int> &iVerts)
{
    indicesBuffer_.bind();
    int nbIndicesBuffer = indicesBuffer_.size() / (sizeof(GLuint)*3);
    indicesBuffer_.release();
    verticesBuffer_.bind();
    int nbVerticesBuffer = verticesBuffer_.size() / (sizeof(GLfloat)*3);
    verticesBuffer_.release();
    int nbTris=iTris.size();
#pragma omp parallel for //recompute triangle normals and aabb
    for (int i=0;i<nbTris;++i)
    {
        Triangle &t=triangles_[iTris[i]];
        const QVector3D &v1=vertices_[t.vIndices_[0]];
        const QVector3D &v2=vertices_[t.vIndices_[1]];
        const QVector3D &v3=vertices_[t.vIndices_[2]];
        t.normal_ = QVector3D::normal(v1,v2,v3);
        t.aabb_ = Geometry::computeTriangleAabb(v1,v2,v3);
    }
    updateOctree(iTris);
    updateNormals(iVerts);
    if(nbIndicesBuffer<getNbTriangles() || nbVerticesBuffer<getNbVertices())
    {
        updateIndexBuffer(iTris);
        updateVertexBuffer(iVerts);
        updateNormalBuffer(iVerts);
    }
    else
        initVBO();
}

/**
 * Update Octree
 * For each triangle we check if its position inside the octree has changed
 * if so... we mark this triangle and we remove it from its former cells
 * We push back the marked triangles into the octree
*/
void Mesh::updateOctree(const std::vector<int> &iTris)
{
    int nbTris = iTris.size();
    std::vector<int> trisToMove;
    for (int i=0;i<nbTris;++i) //recompute position inside the octree
    {
        Triangle &t=triangles_[iTris[i]];
        Octree *leaf=t.leaf_;
        if(!leaf->getAabbSplit().pointInside(t.aabb_.getCenter()))
        {
            trisToMove.push_back(iTris[i]);
            std::vector<int> &trisLeaf = leaf->getTriangles();
            if(trisLeaf.size()>0)
            {
                int iTriLast = trisLeaf.back();
                int iPos = t.posInLeaf_;
                trisLeaf[iPos] = iTriLast;
                triangles_[iTriLast].posInLeaf_ = iPos;
                trisLeaf.pop_back();
            }
        }
        else if(!t.aabb_.isInside(leaf->getAabbLoose()))
            leaf->getAabbLoose().expand(t.aabb_);
    }
    int nbTrisToMove = trisToMove.size();
    for(int i=0;i<nbTrisToMove;++i) //add triangle to the octree
    {
        Triangle &tri = triangles_[trisToMove[i]];
        if(octree_->getAabbLoose().isOutside(tri.aabb_)) //we reconstruct the whole octree, slow... but rare
        {
            Aabb aabb(octree_->getAabbSplit().min_,octree_->getAabbSplit().max_);
            delete octree_;
            QVector3D vecShift = (aabb.max_-aabb.min_)*0.2;
            aabb.min_-=vecShift;
            aabb.max_+=vecShift;
            std::vector<int> triangles;
            for (int i=0;i<getNbTriangles();++i)
                triangles.push_back(i);
            octree_ = new Octree();
            ++Triangle::tagMask_;
            octree_->build(this, triangles, aabb );
            leavesUpdate_.clear();
            break;
        }
        else
        {
            Octree* leaf=tri.leaf_;
            octree_->addTriangle(this,tri);
            if(leaf==tri.leaf_)
            {
                std::vector<int> &trisLeaf = leaf->getTriangles();
                tri.posInLeaf_=trisLeaf.size();
                trisLeaf.push_back(trisToMove[i]);
            }
        }
    }
}

/** Update normals */
void Mesh::updateNormals(const std::vector<int> &iVerts)
{
    int nbVers = iVerts.size();
#pragma omp parallel for
    for (int i=0;i<nbVers;++i)
    {
        Vertex &vert=vertices_[iVerts[i]];
        const std::vector<int> &iTri = vert.tIndices_;
        int nbTri = iTri.size();
        QVector3D normal;
        for (int j=0;j<nbTri;++j)
            normal+=triangles_[iTri[j]].normal_;
        vert.normal_ = normal.normalized();
    }
}

/** Update vertex buffer */
void Mesh::updateVertexBuffer(const std::vector<int> &iVerts)
{
    int nbVerts=iVerts.size();
    GLfloat* verticesArray;
    verticesBuffer_.bind();
    verticesArray = (GLfloat*)verticesBuffer_.map(QGLBuffer::WriteOnly);
#pragma omp parallel for
    for(int i=0;i<nbVerts;++i)
    {
        int j=iVerts[i];
        Vertex &v = vertices_[j];
        verticesArray[j*3] = v.x();
        verticesArray[j*3+1] = v.y();
        verticesArray[j*3+2] = v.z();
    }
    verticesBuffer_.unmap();
    verticesBuffer_.release();
}

/** Update normal buffer */
void Mesh::updateNormalBuffer(const std::vector<int> &iVerts)
{
    int nbVerts=iVerts.size();
    GLfloat* normalsArray;
    normalsBuffer_.bind();
    normalsArray = (GLfloat*)normalsBuffer_.map(QGLBuffer::WriteOnly);
#pragma omp parallel for
    for(int i=0;i<nbVerts;++i)
    {
        int j=iVerts[i];
        QVector3D &n = vertices_[j].normal_;
        normalsArray[j*3] = n.x();
        normalsArray[j*3+1] = n.y();
        normalsArray[j*3+2] = n.z();
    }
    normalsBuffer_.unmap();
    normalsBuffer_.release();
}

/** Update index buffer */
void Mesh::updateIndexBuffer(const std::vector<int> &iTris)
{
    int nbTris = iTris.size();
    GLuint* indicesArray;
    indicesBuffer_.bind();
    indicesArray = (GLuint*)indicesBuffer_.map(QGLBuffer::WriteOnly);
#pragma omp parallel for
    for(int i=0;i<nbTris;++i)
    {
        int j = iTris[i];
        Triangle &tri=triangles_[j];
        indicesArray[j*3]=tri.vIndices_[0];
        indicesArray[j*3+1]=tri.vIndices_[1];
        indicesArray[j*3+2]=tri.vIndices_[2];
    }
    indicesBuffer_.unmap();
    indicesBuffer_.release();
}

/** End of stroke, update octree (cut empty leaves or go deeper if needed) */
void Mesh::checkLeavesUpdate()
{
    Tools::tidy(leavesUpdate_);
    int nbLeaves = leavesUpdate_.size();
    std::vector<Octree*> cutLeaves;
    ++Triangle::tagMask_;
    for(int i=0;i<nbLeaves;++i)
    {
        Octree* leaf = leavesUpdate_[i];
        if(!leaf)
            break;
        if(!leaf->getTriangles().size())
            Octree::checkEmptiness(leaf,cutLeaves);
        else if((int)leaf->getTriangles().size() > Octree::maxTriangles_ && leaf->getDepth() < Octree::maxDepth_)
            leaf->constructCells(this);
    }
    Tools::tidy(cutLeaves);
    int nbCutLeaves = cutLeaves.size();
    for(int i=0;i<nbCutLeaves;++i)
    {
        Octree *oc = cutLeaves[i];
        Octree **child = oc->getChildren();
        for(int j=0;j<8;++j)
            child[j]=0;
    }
    leavesUpdate_.clear();
}

/**
  *****************************************
  *****************************************
  *****************************************
  *****************************************
  ************* UNDO/REDO *****************
  *****************************************
  *****************************************
  *****************************************
  *****************************************
*/

/** Undo setters/Getters */
std::vector<Triangle>& Mesh::getTrianglesState() { return undoIte_->tState_; }
std::vector<Vertex>& Mesh::getVerticesState() { return undoIte_->vState_; }

/** Start push state */
void Mesh::startPushState()
{
    ++Mesh::stateMask_;
    if(beginIte_)
        undo_.clear();
    else if(undo_.size()>10)
        undo_.pop_front();
    beginIte_ = false;
    redo_.clear();
    if(undo_.size())
    {
        std::list<State>::iterator lastIte = undo_.end();
        --lastIte;
        while(undoIte_!=lastIte)
        {
            --lastIte;
            undo_.pop_back();
        }
    }
    undo_.push_back(State());
    undoIte_ = undo_.end();
    --undoIte_;
    undoIte_->nbTrianglesState_ = triangles_.size();
    undoIte_->nbVerticesState_ = vertices_.size();
    undoIte_->aabbState_ = octree_->getAabbSplit();
}

/** Push verts and tris */
void Mesh::pushState(const std::vector<int> &iTris, const std::vector<int> &iVerts)
{
    std::vector<Triangle> &tState = undoIte_->tState_;
    std::vector<Vertex> &vState = undoIte_->vState_;
    int nbTris = iTris.size();
    for(int i=0;i<nbTris;++i)
    {
        Triangle &t = triangles_[iTris[i]];
        if(t.stateFlag_!=Mesh::stateMask_)
        {
            t.stateFlag_ = Mesh::stateMask_;
            tState.push_back(t);
        }
    }
    int nbVerts = iVerts.size();
    for(int i=0;i<nbVerts;++i)
    {
        Vertex &v = vertices_[iVerts[i]];
        if(v.stateFlag_!=Mesh::stateMask_)
        {
            v.stateFlag_ = Mesh::stateMask_;
            vState.push_back(v);
        }
    }
}

/** Undo (also push_back the redo) */
void Mesh::undo()
{
    if(!undo_.size() || beginIte_)
        return;
    State redo;
    int nbTriangles = triangles_.size();
    int nbVertices = vertices_.size();
    redo.nbTrianglesState_ = nbTriangles;
    redo.nbVerticesState_ = nbVertices;
    redo.aabbState_ = octree_->getAabbSplit();
    std::vector<Triangle> &tRedoState = redo.tState_;
    std::vector<Vertex> &vRedoState = redo.vState_;

    int nbTrianglesState  = undoIte_->nbTrianglesState_;
    int nbVerticesState  = undoIte_->nbVerticesState_;
    std::vector<Triangle> &tUndoState = undoIte_->tState_;
    std::vector<Vertex> &vUndoState = undoIte_->vState_;

    int nbTris = tUndoState.size();
    int nbVerts = vUndoState.size();
    //REDO
    if(nbTrianglesState<nbTriangles)
    {
        for(int i=nbTrianglesState;i<nbTriangles;++i) tRedoState.push_back(triangles_[i]);
        for(int i=0;i<nbTris;++i)
        {
            Triangle &t = tUndoState[i];
            if(t.id_<nbTrianglesState) tRedoState.push_back(triangles_[t.id_]);
        }
        triangles_.resize(nbTrianglesState);
    }
    else
    {
        triangles_.resize(nbTrianglesState);
        for(int i=0;i<nbTris;++i)
        {
            Triangle &t = tUndoState[i];
            if(t.id_<nbTriangles) tRedoState.push_back(triangles_[t.id_]);
        }
    }
    if(nbVerticesState<nbVertices)
    {
        for(int i=nbVerticesState;i<nbVertices;++i) vRedoState.push_back(vertices_[i]);
        for(int i=0;i<nbVerts;++i)
        {
            Vertex &v = vUndoState[i];
            if(v.id_<nbVerticesState) vRedoState.push_back(vertices_[v.id_]);
        }
        vertices_.resize(nbVerticesState);
    }
    else
    {
        vertices_.resize(nbVerticesState);
        for(int i=0;i<nbVerts;++i)
        {
            Vertex &v = vUndoState[i];
            if(v.id_<nbVertices) vRedoState.push_back(vertices_[v.id_]);
        }
    }
    //UNDO
    for(int i=0;i<nbTris;++i)
    {
        Triangle &t = tUndoState[i];
        if(t.id_<nbTrianglesState)
            triangles_[t.id_] = t;
    }
    for(int i=0;i<nbVerts;++i)
    {
        Vertex &v = vUndoState[i];
        if(v.id_<nbVerticesState)
            vertices_[v.id_] = v;
    }
    recomputeOctree(undoIte_->aabbState_);
    initVBO();
    redo_.push_back(redo);
    if(undoIte_!=undo_.begin())
    {
        beginIte_ = false;
        --undoIte_;
    }
    else
        beginIte_ = true;
}

/** Redo */
void Mesh::redo()
{
    if(!redo_.size())
        return;
    std::list<State>::iterator redoIte_ = redo_.end();
    --redoIte_;
    int nbTrianglesState  = redoIte_->nbTrianglesState_;
    int nbVerticesState  = redoIte_->nbVerticesState_;
    std::vector<Triangle> &tRedoState = redoIte_->tState_;
    std::vector<Vertex> &vRedoState = redoIte_->vState_;
    triangles_.resize(nbTrianglesState);
    vertices_.resize(nbVerticesState);
    int nbTris = tRedoState.size();
    for(int i=0;i<nbTris;++i)
    {
        Triangle &t = tRedoState[i];
        triangles_[t.id_] = t;
    }
    int nbVerts = vRedoState.size();
    for(int i=0;i<nbVerts;++i)
    {
        Vertex &v = vRedoState[i];
        vertices_[v.id_] = v;
    }
    recomputeOctree(redoIte_->aabbState_);
    initVBO();
    if(!beginIte_)
        ++undoIte_;
    else
        beginIte_ = false;
    redo_.pop_back();
}

/** Recompute octree */
void Mesh::recomputeOctree(const Aabb &aabbSplit)
{
    int nbTriangles = triangles_.size();
    std::vector<int> triangles(nbTriangles);
#pragma omp parallel for
    for (int i=0;i<nbTriangles;++i)
        triangles[i] = i;
    ++Triangle::tagMask_;
    delete octree_;
    octree_ = new Octree();
    octree_->build(this, triangles, aabbSplit);
}
